#!/usr/bin/python

PROGRAM_NAME = 'lazy65.py'
PROGRAM_VERSION = '0.1.0'

import src.tempy as tempy
from src.Symbol import Symbol
from src.Language import Token,TokenReader,TokenType
from src.Assembler import Assembler
from src.packer import *
from src.globals import *
import argparse
import os.path
import sys

asm = None
parser = None

def lineCallback():
	asm.parse()



def main():
	global asm, parser
	ap = argparse.ArgumentParser(prog=PROGRAM_NAME, usage='%(prog)s [options] module')
	ap.add_argument("module", help="Specify input module to read from srcdir")
	ap.add_argument("-s", "--srcdir", help="Specify source directory", default='./')
	ap.add_argument("-o", "--outdir", help="Specify output directory", default='./')
	ap.add_argument("--sym-only", help="Generate symbol file only", action="store_true")
	#ap.add_argument("-v","--verbose", help="", action="store_true")
	ap.add_argument("--version", help="Print version number", action="store_true")
	
	args = ap.parse_args()
	
	if args.version:
		print PROGRAM_NAME + ' V' + PROGRAM_VERSION
		return
	
	path = args.module.replace('.','/')
	modulePath = os.path.normpath(args.srcdir + '/' + path + '.z')
	moduleOutPath = os.path.normpath(args.outdir + '/' + path)
	moduleOutDirs = os.path.dirname(moduleOutPath)
	modulePath1 = modulePath;
	
	if not os.path.exists(modulePath):
		modulePath = os.path.normpath(args.srcdir + '/' + path + '/__root__.z')
		moduleOutDirs = moduleOutPath
		if not os.path.exists(modulePath):
			print 'module not found: '+args.module
			return
	try:
		os.makedirs(moduleOutDirs, 0644);
	except:
		pass
	
	sourceFile = moduleOutPath + '.s'
	headerFile = moduleOutPath + '.h'
	symfile = moduleOutPath + '.sym'
	
	mtime = os.path.getmtime(modulePath)
	
	if not os.path.exists(args.outdir):
		os.makedirs(args.outdir) #TODO: make sub-dirs for sub-modules, too... later on
	parser = tempy.Parser(modulePath)
	asm = Assembler(parser, args.module, args.srcdir, args.outdir, mtime)

	parser.setLineCallback(lineCallback)
	lines = parser.parse() #1st pass
	#''.join(lines)
	with open(symfile,'wb') as f:
		p = Packer()
		asm.globalScope.toObject(p)
		f.write(p.getData())
		f.close()
	#asm.doImports()
	if not args.sym_only:
		parser.lines = lines
		asm.prepare2ndPass()
		lines = parser.parse() #2nd pass
		with open(sourceFile,'w') as f:
			f.write('.include "'+os.path.basename(headerFile)+'"\n')
			for l in lines:
				f.write(l+'\n')
			f.close()
		with open(headerFile,'wb') as f:
			f.write(asm.getHeader())
			f.close()


try:
	main()
except tempy.Error, e:
	if parser == None:
		print 'error: '+e.msg
	else:
		print 'parser error: '+parser.filename+'('+str(parser.lineNumber)+'): ' + e.msg
		print 'near "'+parser.line+'"'
		sys.exit(1)
#cProfile.run('main()')
