from Symbol import Symbol
from Language import *
from globals import *
from packer import *
import os
import tempy

#TODO: if symbol from other module is used in macro, automatically include othermodule.h in header file
#TODO: make referencedModules a dict?

tokenReader = TokenReader()

class Assembler:
	def __init__(self, parser, moduleName, srcdir, outdir, mtime):
		self.p = parser
		self.srcdir = srcdir
		self.outdir = outdir
		self.mtime = mtime
		self.moduleName = moduleName
		self.replaces = []
		self.globalScope = Symbol('*')
		self.importScope = Symbol('*')
		self.globalScope1 = None #symbol tree from 1st pass
		self.scope1 = None #current scope from 1st pass
		self.scope = self.globalScope
		self.nUnnamedSymbols = 0
		self.pass2 = False
		self.inZeropage = False
		self.inMacro = 0
		self.headerLines = []
		self.inEnum = 0
		self.unnamedEnum = False
		self.referencedModules = []
	
	def prepare2ndPass(self):
		self.globalScope1 = self.globalScope
		self.scope1 = self.globalScope1
		self.globalScope = Symbol('*')
		self.scope = self.globalScope
		self.pass2 = True
		self.inZeropage = False
		self.inMacro = 0
		self.nUnnamedSymbols = 0
	
	def getIncludePath(self,path):
		srcpath = self.srcdir+'/'+path
		if os.path.isfile(srcpath):
			return srcpath
		return path
	
	def getFilename(self, moduleName, ext=''):
		path = moduleName.replace('.','/')
		if ext=='sym' or ext=='h' or ext=='s':
			file = os.path.normpath(self.outdir + '/' + path)
		else:
			file = os.path.normpath(self.srcdir + '/' + path)
		file = file.replace('\\','/')
		if ext!='':
			file += '.'+ext
		return file
		
	def getRelativePath(self, path1, path2=None):
		if path2==None:
			path2 = os.path.dirname(self.getFilename(self.moduleName,ext='s'))
		return os.path.relpath(path1, path2).replace('\\','/')
		
	def importSingleModule(self, moduleName, scopeName, generateInclude=False):
		if generateInclude:
			self.p.insertLine('.include "'+self.getRelativePath(self.getFilename(moduleName,'h'))+'"')
		
		symfile = self.getFilename(moduleName, 'sym')
		if os.path.isfile(symfile):
			symfileMTime = os.path.getmtime(symfile)
		else:
			symfileMTime = None
		
		if symfileMTime==None or symfileMTime>self.mtime:
			command = PROGRAM_EXEC + ' --sym-only -s "'+self.srcdir+'" -o "'+self.outdir+'" '+moduleName
			res = os.system(command)
			if res!=0:
				raise tempy.Error('Could not create symbol file for module '+moduleName+': Error '+str(res))
		
		#add symbols to scope
		symfile = self.getFilename(moduleName, 'sym')
		with open(symfile,'rb') as f:
			data = f.read()
			u = Unpacker(data)
			name = u.dstring()
			scopeNameList = scopeName.split('.')
			lastScopeName = scopeNameList[-1]
			scope = Symbol(lastScopeName,'','',True,moduleName).fromObject(u)
			if lastScopeName=='*':
				for name in scope.symbols:
					s = scope.symbols[name]
					self.importScope.addSymbol(s)
					s.moduleName = moduleName
					#s.globalName = '*.' + moduleName + s.globalName[1:]
					s.imported = True
			else:
				#create import scope
				currentScope = self.importScope
				if len(scopeNameList)>1:
					for name in scopeNameList[:-1]:
						s = currentScope.getSymbol([name])
						if s==None:
							s = Symbol(name,'','',True)
							currentScope.addSymbol(s)
						currentScope = s
				currentScope.addSymbol(scope)
			#self.importScope.printTree('    ')
	
	def importModule(self, moduleName, scopeName, importSubModules=True, generateIncludes=True):
		modulePath = self.srcdir + '/' + moduleName.replace('.','/')
		if os.path.isfile(modulePath+'.z') or os.path.isfile(modulePath+'/__root__.z'):
			self.importSingleModule(moduleName, scopeName, generateIncludes)
		
		path = self.getFilename(moduleName)
		if importSubModules and os.path.isdir(path):
			inRootFolder = True
			for root, dirs, files in os.walk(path):
				if self.srcdir!='':
					root = root[len(self.srcdir)+1:]
				root = root.replace('\\','.').replace('/','.')
				#path = root.split('.')
				#print (len(path) - 1) *'   ' , '+', os.path.basename(root)
				files = sorted(files,key=str.lower) #make sure, __root__.z comes first
				for file in files:
					ext = os.path.splitext(file)[1][1:]
					if ext=='z' or ext=='Z':
						file = file[:-2]
						if inRootFolder and file=='__root__':
							pass
						else:
							subModuleName = ''
							if file=='__root__':
								subModuleName = root
							else:
								subModuleName = root + '.' + file
							if scopeName=='*':
								subModuleScopeName = ''
							else:
								subModuleScopeName = scopeName+'.'
							subModuleScopeName += subModuleName[len(moduleName)+1:]
							self.importSingleModule(subModuleName, subModuleScopeName, generateIncludes)
							#print len(path)*'   ', ' ', subModuleName, subModuleScopeName
				inRootFolder = False
		
	def convertName(self, s):
		name = s.globalName
		if name[0]=='*':
			if s.imported:
				out = '_'+s.moduleName.replace('.','_')
				if s.moduleName not in self.referencedModules:
					self.referencedModules.append(s.moduleName)
			else:
				out = '_'+self.moduleName.replace('.','_')
			for c in name[1:]:
				if c=='_':
					out += '__'
				elif c=='.':
					out += '_'
				else:
					out += c
			return out
		return name
	
	def replace(self,text,rep):
		offs = 0
		for r in rep:
			start = r[0] + offs
			end = r[1] + offs
			replaceText = r[2]
			length = end - start
			offs += len(replaceText) - length
			text = text[:start] + replaceText + text[end:]
		return text
	
	def parse(self):
		#first pass: populate symbol tree
		text = self.p.line
		tokenReader.reset(text)
		tok = tokenReader.read()
		if tok != None:
			if tok.type == TokenType.COMMAND:
				self.command(tok.text)
			elif tok.type == TokenType.LABEL:
				if tok.text[0]!=LOCAL_LABEL_CHAR:
					self.addSymbol(Symbol(tok.text,'label'))
			elif tok.type == TokenType.ID:
				if not self.pass2:
					#assignment - 1st pass
					tok2 = tokenReader.read()
					if tok2!=None and tok2.text=='=':
						tok3 = tokenReader.readRest()
						self.addSymbol(Symbol(tok.text,'constant',tok3.text))
		if self.pass2:
			self.parse2nd()
	
	def parse2nd(self):
		#second pass: replace symbol references
		line = self.p.line
		tokenReader.reset(line)
		self.replaces = []
		prevTok = None
		isAssignment = False
		assignmentLeft = ''
		self.referencedModules = []
		while tokenReader.pos<len(line):
			tok = tokenReader.read()
			if tok!=None:
				if tok.type==TokenType.COMMENT:
					break
				elif tok.type==TokenType.COMMAND:
					#convert scopes to labels
					cmd = tok.text
					if cmd in SCOPE_COMMANDS:
						tok2 = tokenReader.read()
						symbolName = ''
						if tok2==None: #unnamed
							#symbolName = '__symbol'+str(self.nUnnamedSymbols)+'__'
							line = ''
							continue #no label needed for unnamed symbol
						else:
							symbolName = tok2.text
						s = self.getSymbol(symbolName)
						if s==None:
							raise tempy.Error('Unknown symbol "'+symbolName+'"')
						#check if scope has been defined previously in this pass
						#do not emit another label if so
						if self.scope.parent==None:
							raise tempy.Error('Oops, this should not happen!')
						sPrev = self.scope.parent.getSymbol([symbolName])
						if sPrev==None:
							raise tempy.Error('Oops, this should not happen!')
						else:
							if sPrev.nDeclarations>1:
								line = ''
							else:
								line = self.convertName(s)+':'
						break
					elif cmd in SCOPE_CLOSE_COMMANDS:
						line = ''
						break
				elif tok.type==TokenType.ID:
					#replace symbol references
					s = self.getSymbol(tok.text)
					if s != None:
						self.addReplace(tokenReader.startPos, tokenReader.pos, self.convertName(s))
				elif tok.type==TokenType.SPECIAL:
					if tok.text=='=' and prevTok!=None and prevTok.type==TokenType.ID:
						isAssignment = True
						assignmentLeft = prevTok.text
				elif tok.type==TokenType.LABEL:
					if tok.text[0]!='@':
						s = self.scope.getSymbol([tok.text])
						self.addReplace(tokenReader.startPos, tokenReader.pos, self.convertName(s)+':')
			prevTok = tok
		line = self.replace(line, self.replaces)
		
		if isAssignment:
			i = line.index('=')
			for m in self.referencedModules:
				self.addHeaderLine('.include "'+self.getRelativePath(self.getFilename(m,'h'))+'"')
			self.addSymbol(Symbol(assignmentLeft,'constant',line[i+1:]))
			line = ''
		
		if self.inMacro > 0:
			self.addHeaderLine(line)
			if self.inMacro==1:
				self.inMacro = 0 #include .endmacro
			#self.p.discardLine()
			self.p.line = '' #just clear line so line numbers of reported errors match between file.z/file.s
		elif self.inEnum > 0:
			if self.inEnum==1:
				self.inEnum = 0 #include .endenum
			#self.p.discardLine()
			self.p.line = ''
		else:
			self.p.line = line
	
	def getSymbol(self, name):
		if name[:2]=='::':
			nameList = splitScopes(name[2:])
			s = self.globalScope1.getSymbol(nameList)
			if s==None:
				s = self.importScope.getSymbol(nameList)
			return s
		else:
			nameList = splitScopes(name)
			s = self.scope1.getSymbol(nameList,True)
			if s==None:
				s = self.importScope.getSymbol(nameList)
			return s
	
	def addReplace(self, start, end, text):
		self.replaces.append((start, end, text))
	
	def command(self,cmd):
		if cmd in SCOPE_COMMANDS:
			tok = tokenReader.read()
			symbolName = ''
			if(tok == None):
				symbolName = '__symbol'+str(self.nUnnamedSymbols)+'__'
				s = self.addSymbol(Symbol(symbolName,cmd))
				self.setCurrentScope(s)
				self.nUnnamedSymbols += 1
			elif(tok.type == TokenType.ID):
				symbolName = tok.text
				s = self.mergeSymbol(Symbol(symbolName,cmd))
				self.setCurrentScope(s)
			else:
				raise tempy.Error('identifier expected')
		elif cmd in SCOPE_CLOSE_COMMANDS:
			if self.scope.type == cmd[3:]:
				if self.scope.parent!=None:
					self.setCurrentScope(self.scope.parent)
				else:
					raise tempy.Error('cannot close scope, no parent')
			else:
				raise tempy.Error('tried to close "'+self.scope.type+'" with .'+cmd)
				
		elif cmd in MACRO_COMMANDS: #treat macros as regular scopes
			self.inMacro = 2
			tok = tokenReader.read()
			if(tok.type == TokenType.ID):
				s = self.addSymbol(Symbol(tok.text,'macro'))
				self.setCurrentScope(s)
			else:
				raise tempy.Error('identifier expected')
			#add macro args to local symbol list
			args = self.readArgList()
			for a in args:
				if len(a)==1 and a[0].type==TokenType.ID:
					s.addSymbol(Symbol(a[0].text))
				elif len(a)>1:
					raise tempy.Error('too many tokens in macro argument')
		elif cmd in MACRO_CLOSE_COMMANDS:
			if self.scope.parent!=None:
				self.setCurrentScope(self.scope.parent)
			else:
				raise tempy.Error('cannot close macro scope, no parent')
			self.inMacro = 1
			
		elif cmd == 'enum':
			if self.inEnum > 0:
				tempy.Error('nested enums are not supported')
			tok = tokenReader.read()
			if tok==None or tok.type==TokenType.COMMENT:
				#unnamed enum - symbols remain in current scope
				self.unnamedEnum = True
			elif tok.type==TokenType.ID:
				s = self.addSymbol(Symbol(tok.text,cmd), False) #root symbol of enum scope is not referenceable
				self.setCurrentScope(s)
				self.unnamedEnum = False
			else:
				raise tempy.Error('identifier expected')
			self.inEnum = 2
		elif cmd == 'endenum':
			self.inEnum = 1
			if not self.unnamedEnum:
				#back to parent scope
				if self.scope.parent!=None:
					self.setCurrentScope(self.scope.parent)
				else:
					raise tempy.Error('cannot close enum, no parent')

		elif self.pass2:
			if cmd=='use': #.use module[, importScope]
				tok = tokenReader.read()
				if tok==None:
					raise tempy.Error('module name expected')
				moduleName = tok.text
				tok = tokenReader.read()
				scopeName = moduleName
				if tok!=None:
					if tok.text!=',':
						raise tempy.Error('comma expected')
					tok = tokenReader.read()
					if tok!=None and tok.text=='*':
						scopeName = '*'
					elif tok!=None and tok.type==TokenType.ID:
						scopeName = tok.text
					else:
						raise tempy.Error('identifier or "*" expected')
				self.importModule(moduleName, scopeName, True, True)
				self.p.discardLine()
			elif cmd=='segment':
				tok = tokenReader.read()
				if tok==None:
					raise tempy.Error('segment name expected')
				self.inZeropage = (tok.text=='ZEROPAGE')
			elif cmd=='zeropage':
				self.inZeropage = True
			elif cmd in ('code','bss','data','rodata'):
				self.inZeropage = False
			elif cmd=='incbin' or cmd=='include':
				tok = tokenReader.read()
				if tok==None:
					raise tempy.Error('filename expected')
				path = self.getIncludePath(tok.text)
				self.p.line = '.'+cmd+' "'+path+'"' + tokenReader.readRest().text
		
	def readArgList(self):
		i = 0
		args = [[]]
		while True:
			tok = tokenReader.read()
			if tok==None or tok.type==TokenType.COMMENT:
				break
			elif tok.type==TokenType.SPECIAL:
				if tok.text==',':
					i+=1
					args.append([])
			else:
				args[i].append(tok)
		return args
	
	def addHeaderLine(self, line):
		self.headerLines.append(line)
	
	def addHeaderSymbol(self, s):
		line = ''
		if s.value!='': #is constant?
			line += self.convertName(s) + ' ='
			line += s.value
		else:
			if self.inZeropage:
				line += '.globalzp '
			else:
				line += '.global '
			line += self.convertName(s)
		self.headerLines.append(line)
	
	def addSymbol(self,s,addToHeader=True):
		s = self.scope.addSymbol(s)
		if self.pass2 and addToHeader and s.type!='macro':
			self.addHeaderSymbol(s)
		return s

	def mergeSymbol(self,s):
		if(self.scope.getSymbol([s.name]) != None):
			#symbol has been added previously, don't add to header again
			return self.scope.mergeSymbol(s)
		else:
			return self.addSymbol(s)
	
	def setCurrentScope(self, symbol):
		self.scope = symbol
		if self.globalScope1 != None:
			if symbol.name=='*':
				self.scope1 = self.globalScope1
			else:
				nameList = splitScopes(symbol.globalName[2:])
				self.scope1 = self.globalScope1.getSymbol(nameList)
				if self.scope1 == None:
					raise tempy.Error('Unknown symbol "'+symbol.name+'"')
	
	def getHeader(self):
		convertedModuleName = self.moduleName.replace('.','_').upper()
		out = '.ifndef __'+convertedModuleName+'_H\n'
		out += '__'+convertedModuleName+'_H = 1\n'
		return out+('\n'.join(self.headerLines))+'\n.endif\n'
	
