
class Symbol:
	def __init__(self,name,type='',value='',imported=False,moduleName=''):
		self.moduleName = moduleName
		self.name = name
		self.globalName = name
		self.value = value
		self.parent = None
		self.type = type
		self.imported = imported
		self.symbols = {}
		self.nDeclarations = 1
	
	def getValue(self):
		return value
	
	def addSymbol(self, symbol):
		symbol.parent = self
		symbol.imported = self.imported
		symbol.moduleName = self.moduleName
		symbol.globalName = self.globalName + '.' + symbol.name
		self.symbols[symbol.name] = symbol
		return symbol
	
	def mergeSymbol(self, symbol):
		#used to allow multiple scope definitions with the same name per module
		if symbol.name in self.symbols:
			self.symbols[symbol.name].symbols.update(symbol.symbols)
			self.symbols[symbol.name].nDeclarations += symbol.nDeclarations
			return self.symbols[symbol.name]
		else:
			return self.addSymbol(symbol)
	
	def getSymbol(self, nameList, searchParents=False):
		if len(nameList)>0:
			name = nameList[0]
			if name in self.symbols:
				if 1==len(nameList):
					return self.symbols[name]
				return self.symbols[name].getSymbol(nameList[1:])
			elif searchParents==False and name==self.name:
				return self
		if searchParents and self.parent!=None:
			return self.parent.getSymbol(nameList,True)
		return None
	
	def toObject(self, p):
		includedSymbols = self.symbols
		p.dstring(self.name)
		p.dstring(self.value)
		p.u32(len(includedSymbols))
		#p.u8(ord('\n')) #debug
		for name in includedSymbols:
			s = includedSymbols[name]
			s.toObject(p)
			#p.u8(ord('\n')) #debug
	
	def fromObject(self, u, parent=None): #self.name = u.dstring()
		self.value = u.dstring()
		nSymbols = u.u32()
		self.parent = parent
		if parent!=None:
			self.moduleName = parent.moduleName
			self.imported = parent.imported
			self.globalName = parent.globalName + '.' + self.name
		else:
			self.globalName = '*'
		for i in range(nSymbols):
			name = u.dstring()
			s = Symbol(name).fromObject(u,self)
			s.imported = self.imported
			self.addSymbol( s )
			#u.u8() #debug
		return self

	def printTree(self, indent=''): #debug
		for name in self.symbols:
			print indent + name
			self.symbols[name].printTree(indent+'    ')
