import struct

#currently optimized for ouput size

class Packer:
	def __init__(self):
		self.data = ''
		self.st_s32 = struct.Struct('<l')
		self.st_u32 = struct.Struct('<L')
		self.st_s64 = struct.Struct('<q')
		self.st_double = struct.Struct('<d')
	
	def getData(self):
		return self.data
	
	def u8(self, v):
		self.data += chr(v)
		
	def s32(self, v):
		self.data += self.st_s32.pack(v)
		
	def u32(self, v):
		self.data += self.st_u32.pack(v)
	
	def s64(self, v):
		self.data += self.st_s64.pack(v)
		
	def double(self, v):
		self.data += self.st_double.pack(v)
		
	def fstring(self, v, n):
		for i in range(n):
			if i>=len(v):
				self.data += '\0'
			else:
				self.data += v[i]
	
	def dstring(self, v, delimiter='\0'): #variable-length, (zero-)delimited
		if not isinstance(v, str):
			raise Exception('string expected')
		if v.find(delimiter)!=-1:
			raise Exception("string must not contain delimiter")
		self.data += v + delimiter
		
	def string(self, v): #variable-length, length specified by uint
		if not isinstance(v, str):
			raise Exception('string expected')
		self.u32(len(v))
		self.fstring(v, len(v))
	
class Unpacker:
	def __init__(self, data):
		self.data = data
		self.pos = 0
		self.st_s32 = struct.Struct('<l')
		self.st_u32 = struct.Struct('<L')
		self.st_s64 = struct.Struct('<q')
		self.st_double = struct.Struct('<d')
	
	def u8(self):
		v = ord(self.data[self.pos])
		self.pos += 1
		return v
		
	def s32(self):
		v = self.st_s32.unpack_from(self.data, self.pos)[0]
		self.pos += self.st_s32.size
		return v
		
	def u32(self):
		v = self.st_u32.unpack_from(self.data, self.pos)[0]
		self.pos += self.st_u32.size
		return v
	
	def s64(self):
		v = self.st_s64.unpack_from(self.data, self.pos)[0]
		self.pos += self.st_s64.size
		return v
		
	def double(self):
		v = self.st_double.unpack_from(self.data, self.pos)[0]
		self.pos += self.st_double.size
		return v
		
	def fstring(self, n):
		v = self.data[self.pos:self.pos+n]
		self.pos += n
		return v
	
	def dstring(self, delimiter='\0'):
		i = self.data[self.pos:].find(delimiter)
		if i!=-1:
			start = self.pos
			self.pos += i+1
			return self.data[start:self.pos-1]
		else:
			raise Exception('string not delimited')
			
	def string(self):
		return self.fstring( self.u32() )
		